const Ast = @import("./kym/Ast.zig");

const coral = @import("coral");

const file = @import("./file.zig");

const tokens = @import("./kym/tokens.zig");

///
///
///
pub const CallContext = struct {
	env: *RuntimeEnv,
	obj: Value,
	args: []const Value = &.{},

	///
	///
	///
	pub fn arg_at(self: CallContext, index: Int) RuntimeError!Value {
		if (!coral.math.is_clamped(index, 0, self.args.len - 1)) {
			return self.env.check_fail("argument out of bounds");
		}

		return self.args[@intCast(usize, index)];
	}
};

const Chunk = struct {
	env: *RuntimeEnv,
	opcodes: coral.list.Stack(Opcode),

	///
	///
	///
	const Opcode = union (enum) {
		push_nil,
		push_true,
		push_false,
		push_zero,
		push_number: Float,
		push_array: i32,
		push_table: i32,

		not,
		neg,

		add,
		sub,
		mul,
		div,

		eql,
		cgt,
		clt,
		cge,
		cle,
	};

	pub fn compile_ast(self: *Chunk, ast: Ast) void {
		for (ast.list_statements()) |statement| {
			switch (statement) {
				.return_expression => |return_expression| {
					try self.compile_expression(return_expression);
				},

				.return_nothing => {
					try self.emit(.push_nil);
				},
			}
		}
	}

	pub fn compile_expression(self: *Chunk, expression: Ast.Expression) void {
		switch (expression) {
			.nil_literal => try self.emit(.push_nil),
			.true_literal => try self.emit(.push_true),
			.false_literal => try self.emit(.push_false),
			.integer_literal => |literal| try self.emit(if (literal == 0) .push_zero else .{.push_integer = literal}),
			.float_literal => |literal| try self.emit(if (literal == 0) .push_zero else .{.push_float = literal}),
			.string_literal => |literal| try self.emit(.{.push_object = try self.intern(literal)}),

			.array_literal => |elements| {
				if (elements.values.len > coral.math.max_int(@typeInfo(Int).Int)) {
					return error.OutOfMemory;
				}

				for (elements.values) |element_expression| {
					try self.compile_expression(element_expression);
				}

				try self.emit(.{.push_array = @intCast(Int, elements.values.len)});
			},

			.table_literal => |fields| {
				if (fields.values.len > coral.math.max_int(@typeInfo(Int).Int)) {
					return error.OutOfMemory;
				}

				for (fields.values) |field| {
					try self.compile_expression(field.expression);
					try self.emit(.{.push_object = try self.intern(field.identifier)});
				}

				try self.emit(.{.push_table = @intCast(Int, fields.values.len)});
			},

			.binary_operation => |operation| {
				try self.compile_expression(operation.lhs_expression.*);
				try self.compile_expression(operation.rhs_expression.*);

				try self.emit(switch (operation.operator) {
					.addition => .add,
					.subtraction => .sub,
					.multiplication => .mul,
					.divsion => .div,
					.greater_equals_comparison => .compare_eq,
					.greater_than_comparison => .compare_gt,
					.equals_comparison => .compare_ge,
					.less_than_comparison => .compare_lt,
					.less_equals_comparison => .compare_le,
				});
			},

			.unary_operation => |operation| {
				try self.compile_expression(operation.expression.*);

				try self.emit(switch (operation.operator) {
					.boolean_negation => .not,
					.numeric_negation => .neg,
				});
			},

			.grouped_expression => |grouped_expression| {
				try self.compile_expression(grouped_expression.*);
			}
		}
	}

	pub fn deinit(self: *Chunk) void {
		self.opcodes.deinit(self.env.allocator);
	}

	pub fn execute(self: *Chunk) RuntimeError!Value {
		_ = self;

		return Value.nil;
	}
};

pub const Float = f64;

pub const Int = i32;

pub const Objectid = u32;

pub const RuntimeEnv = struct {
	output: coral.io.Writer,
	stack: coral.list.Stack(u64),
	interned: coral.table.Hashed([]const u8, Objectid, coral.table.string_keyer),
	objects: coral.slab.Map(@typeInfo(u32).Int, Object),
	user_allocator: coral.io.Allocator,

	pub const DataSource = struct {
		name: []const u8,
		data: []const u8,
	};

	pub const Limits = struct {
		stack_max: u32,
		calls_max: u32,
	};

	const Object = struct {
		ref_count: usize,

		state: struct {
			info: ObjectInfo,
			userdata: []u8,

			fields: coral.table.Hashed(*Object, *Value, .{
				.hasher = struct {
					fn hash(object: *Object) coral.table.Hash {
						coral.debug.assert(object.state.info.identity == null);

						return coral.table.hash_string(object.state.userdata);
					}
				}.hash,

				.comparer = struct {
					fn compare(object_a: *Object, object_b: *Object) isize {
						coral.debug.assert(object_a.state.info.identity == null);
						coral.debug.assert(object_b.state.info.identity == null);

						return coral.io.compare(object_a.state.userdata, object_b.state.userdata);
					}
				}.compare,
			}),
		},
	};

	pub const ObjectInfo = struct {
		caller: *const fn (caller: Value, context: CallContext) RuntimeError!Value = default_call,
		cleaner: *const fn (userdata: []u8) void = default_clean,
		getter: *const fn (context: CallContext) RuntimeError!Value = default_get,
		identity: ?*const anyopaque = null,
		setter: *const fn (context: CallContext) RuntimeError!void = default_set,

		fn default_call(_: Value, context: CallContext) RuntimeError!Value {
			return context.env.fail(error.BadOperation, "attempt to call non-callable");
		}

		fn default_clean(_: []u8) void {
			// Nothing to clean up by default.
		}

		fn default_get(context: CallContext) RuntimeError!Value {
			return context.env.fail(error.BadOperation, "attempt to get non-indexable");
		}

		fn default_set(context: CallContext) RuntimeError!void {
			return context.env.fail(error.BadOperation, "attempt to set non-indexable");
		}
	};

	pub fn call(self: *RuntimeEnv, caller: Value, maybe_index: ?Value, args: []const Value) RuntimeError!RuntimeVar {
		if (maybe_index) |index| {
			const callable = try self.get(caller, index);

			defer callable.deinit();

			return switch (callable.value.unpack()) {
				.objectid => |callable_id| .{
					.env = self,

					.value = try self.objects.fetch(callable_id).state.info.caller(.{
						.env = self,
						.callable = callable.value,
						.obj = caller,
						.args = args,
					}),
				},

				else => self.fail(error.BadOperation, "attempt to call non-object type"),
			};
		}

		return self.bind(try self.objects.fetch(try self.to_objectid(caller)).state.info.caller(.{
			.env = self,
			.obj = caller,
			// .caller = .{.object = self.global_object},
			.args = args,
		}));
	}

	pub fn check(
		self: *RuntimeEnv,
		condition: bool,
		runtime_error: RuntimeError,
		failure_message: []const u8) RuntimeError!void {

		if (condition) {
			return;
		}

		return self.fail(runtime_error, failure_message);
	}

	pub fn deinit(self: *RuntimeEnv) void {
		self.stack.deinit();
	}

	pub fn execute_data(self: *RuntimeEnv, allocator: coral.io.Allocator, source: DataSource) RuntimeError!RuntimeVar {
		var ast = try Ast.init(allocator);

		defer ast.deinit();

		{
			var tokenizer = tokens.Tokenizer{.source = source.data};

			try ast.parse(&tokenizer);
		}

		var chunk = Chunk{
			.env = self,
			.opcodes = .{.allocator = allocator},
		};

		const typeid = "<chunk>";

		const script = try self.new_object(allocator, coral.io.bytes_of(&chunk), .{
			.identity = typeid,

			.cleaner = struct {
				fn clean(userdata: []const u8) void {
					@ptrCast(*Chunk, @alignCast(@alignOf(Chunk), userdata)).deinit();
				}
			}.clean,

			.caller = struct {
				fn call(caller: Value, context: CallContext) RuntimeError!Value {
					_ = caller;

					return (context.env.native_cast(context.obj, typeid, Chunk) catch unreachable).execute();
				}
			}.call,
		});

		defer script.deinit();

		return try self.call(script.value, null, &.{});
	}

	pub fn execute_file(
		self: *RuntimeEnv,
		allocator: coral.io.Allocator,
		file_system: file.System,
		file_path: file.Path) RuntimeError!RuntimeVar {

		const readable_file = file_system.open_readable(file_path) catch return error.SystemFailure;

		defer readable_file.close();

		var file_data = coral.list.ByteStack{.allocator = allocator};
		const file_size = (file_system.query_info(file_path) catch return error.SystemFailure).size;

		try file_data.grow(file_size);

		defer file_data.deinit();

		{
			var stream_buffer = [_]u8{0} ** 4096;

			if ((coral.io.stream(coral.list.stack_as_writer(&file_data), readable_file.as_reader(), &stream_buffer) catch {
				return error.SystemFailure;
			}) != file_size) {
				return error.SystemFailure;
			}
		}

		return try self.execute_data(allocator, .{
			.name = file_path.to_string() catch return error.SystemFailure,
			.data = file_data.values,
		});
	}

	pub fn fail(self: *RuntimeEnv, runtime_error: RuntimeError, failure_message: []const u8) RuntimeError {
		// TODO: Call stack and line numbers.
		coral.utf8.print_formatted(self.output, "{name}@({line}): {message}\n", .{
			.name = ".ona",
			.line = @as(u64, 0),
			.message = failure_message,
		}) catch return error.SystemFailure;

		return runtime_error;
	}

	pub fn get(self: *RuntimeEnv, indexable: Value, index: Value) RuntimeError!RuntimeVar {
		const indexable_object = self.objects.fetch(switch (indexable.unpack()) {
			.objectid => |indexable_id| indexable_id,
			else => return self.fail(error.BadOperation, "attempt to index non-indexable type"),
		});

		return .{
			.env = self,

			.value = try indexable_object.state.info.getter(.{
				.env = self,
				.obj = indexable,
				.args = &.{index},
			}),
		};
	}

	pub fn init(allocator: coral.io.Allocator, output: coral.io.Writer, limits: Limits) RuntimeError!RuntimeEnv {
		var env = RuntimeEnv{
			.output = output,
			.stack = .{.allocator = allocator},
			.objects = .{.allocator = allocator},
			.interned = .{.allocator = allocator},
			.user_allocator = allocator,
		};

		try env.stack.grow(limits.stack_max * limits.calls_max);

		errdefer env.stack.deinit();

		return env;
	}

	pub fn intern(self: *RuntimeEnv, string: []const u8) RuntimeError!Value {
		return Value.pack_objectid(self.interned.lookup(string) orelse {
			const interned_value = (try self.new_string(string)).value;

			switch (interned_value.unpack()) {
				.objectid => |id| coral.debug.assert(try self.interned.insert(string, id)),
				else => unreachable,
			}

			return interned_value;
		});
	}

	pub fn native_cast(self: *RuntimeEnv, castable: Value, id: *const anyopaque, comptime Type: type) RuntimeError!*Type {
		const object = self.objects.fetch(castable.to_objectid() orelse {
			return self.fail(error.BadOperation, "attempt to cast non-castable type");
		});

		const is_expected_type = (object.state.info.identity == id) and (object.state.userdata.len == @sizeOf(Type));

		try self.check(is_expected_type, "invalid object cast: native type");

		return @ptrCast(*Type, @alignCast(@alignOf(Type), object.state.userdata));
	}

	pub fn new_object(self: *RuntimeEnv, allocator: coral.io.Allocator, userdata: []const u8, info: ObjectInfo) RuntimeError!RuntimeVar {
		const allocation = try coral.io.allocate_many(allocator, userdata.len, u8);

		errdefer coral.io.deallocate(allocator, allocation);

		coral.io.copy(allocation, userdata);

		const objectid = try self.objects.insert(.{
			.ref_count = 1,

			.state = .{
				.info = info,
				.userdata = allocation,
				.fields = .{.allocator = allocator},
			},
		});

		return .{
			.env = self,
			.value = .{.data = @as(usize, 0x7FF8000000000002) | (@as(usize, objectid) << 32)},
		};
	}

	pub fn new_string(self: *RuntimeEnv, data: []const u8) RuntimeError!RuntimeVar {
		return try self.new_object(data, .{
			.getter = struct {
				fn get_byte(context: CallContext) RuntimeError!*Value {
					const string = context.env.string_cast(context.obj) catch unreachable;
					const index = try context.env.to_int(try context.arg_at(0));

					try context.env.check(coral.math.is_clamped(index, 0, string.len), "index out of string bounds");

					return context.env.new_int(string[@intCast(usize, index)]);
				}
			}.get_byte,
		});
	}
};

pub const RuntimeError = coral.io.AllocationError || Ast.ParseError || error {
	BadOperation,
	BadArgument,
	SystemFailure,
};

pub const RuntimeVar = struct {
	value: Value,
	env: *RuntimeEnv,

	pub fn deinit(self: RuntimeVar) void {
		switch (self.value.unpack()) {
			.objectid => |id| {
				const object = self.env.objects.fetch(id);

				coral.debug.assert(object.ref_count != 0);

				object.ref_count -= 1;

				if (object.ref_count == 0) {
					object.state.info.cleaner(object.state.userdata);
					// TODO: Free individual key-value pairs of fields
					object.state.fields.deinit();
					coral.io.deallocate(self.env.user_allocator, object.state.userdata);
					self.env.objects.remove(id);
				}
			},

			else => {},
		}
	}
};

pub const Value = struct {
	data: u64,

	pub const Unpacked = union (enum) {
		nil,
		false,
		true,
		number: Float,
		objectid: Objectid,
	};

	fn pack_number(float: Float) Value {
		return @bitCast(Value, float);
	}

	fn pack_objectid(id: Objectid) Value {
		return signature.objectid | id;
	}

	pub const @"false" = @as(Value, nan | 0x0001000000000000);

	const mask = .{
		.sign = @as(u64, 0x8000000000000000),
		.exponent = @as(u64, 0x7ff0000000000000),
		.quiet = @as(u64, 0x0008000000000000),
		.type = @as(u64, 0x0007000000000000),
		.signature = @as(u64, 0xffff000000000000),
		.object = @as(u64, 0x00000000ffffffff),
	};

	pub const nan = @as(Value, mask.exponent | mask.quiet);

	pub const nil = @as(Value, nan | 0x0003000000000000);

	const signature = .{

	};

	pub const @"true" = @as(Value, nan | 0x0002000000000000);

	pub fn unpack(self: Value) Unpacked {
		if ((~self.data & mask.exponent) != 0) {
			return .{.number = @bitCast(Float, self.data)};
		}

		return switch ((self.data & mask.signature) != 0) {
			.signature_nan => .{.number = @bitCast(Float, self.data)},
			.signature_false => .false,
			.signature_true => .true,
			.signature_object => @intCast(Objectid, self.data & mask.object),
			else => return .nil,
		};
	}
};
