const builtin = @import("builtin");

const canvas = @import("./canvas.zig");

const coral = @import("coral");

const ext = @import("./ext.zig");

pub const file = @import("./file.zig");

pub const heap = @import("./heap.zig");

const kym = @import("./kym.zig");

const AppManifest = struct {
	title: [255:0]u8 = [_:0]u8{0} ** 255,
	width: u16 = 640,
	height: u16 = 480,

	pub fn load_script(self: *AppManifest, env: *kym.RuntimeEnv, fs: file.System, file_path: []const u8) !void {
		var manifest = try env.execute_file(heap.allocator, fs, file.Path.from(&.{file_path}));

		defer manifest.deinit();

		{
			var title = try env.get(manifest.value, try env.intern("title"));

			defer title.deinit();

			const title_string = try env.string_cast(title.value);

			try env.check(title_string.len <= self.title.len, "`title` cannot exceed 255 bytes in length");
			coral.io.copy(&self.title, title_string);
		}

		const u16_max = coral.math.max_int(@typeInfo(u16).Int);

		{
			const width = try env.get(manifest.value, try env.intern("width"));

			errdefer width.deinit();

			if (width.value.as_number()) |value| {
				if (value < u16_max) {
					self.width = @floatToInt(u16, value);
				}
			}
		}

		{
			const height = try env.get(manifest.value, try env.intern("height"));

			errdefer height.deinit();

			if (height.value.as_number()) |value| {
				if (value < u16_max) {
					self.height = @floatToInt(u16, value);
				}
			}
		}
	}
};

fn stack_as_log_writer(self: *coral.list.ByteStack) coral.io.Writer {
	return coral.io.Writer.bind(coral.list.ByteStack, self, struct {
		fn write(stack: *coral.list.ByteStack, bytes: []const coral.io.Byte) ?usize {
			var line_written = @as(usize, 0);

			for (bytes) |byte| {
				if (byte == '\n') {
					ext.SDL_LogError(ext.SDL_LOG_CATEGORY_APPLICATION, "%.*s", stack.values.len, stack.values.ptr);
					stack.clear();

					line_written = 0;

					continue;
				}

				stack.push_one(byte) catch {
					coral.debug.assert(stack.drop(line_written));

					return null;
				};

				line_written += 1;
			}

			return bytes.len;
		}
	}.write);
}

pub fn run_app(base_file_system: file.System) void {
	defer heap.trace_leaks();

	var log_buffer = coral.list.ByteStack{.allocator = heap.allocator};

	defer log_buffer.deinit();

	var script_env = kym.RuntimeEnv.init(heap.allocator, stack_as_log_writer(&log_buffer), .{
		.stack_max = 512,
		.calls_max = 512,
	}) catch {
		return ext.SDL_LogError(ext.SDL_LOG_CATEGORY_APPLICATION, "failed to initialize Kym vm\n");
	};

	defer script_env.deinit();

	const app_file_name = "app.ona";
	var app_manifest = AppManifest{};

	app_manifest.load_script(&script_env, base_file_system, app_file_name) catch {
		return ext.SDL_LogError(ext.SDL_LOG_CATEGORY_APPLICATION, "failed to load %s\n", app_file_name);
	};

	if (ext.SDL_Init(ext.SDL_INIT_EVERYTHING) != 0) {
		return ext.SDL_LogError(ext.SDL_LOG_CATEGORY_APPLICATION, "%s\n", ext.SDL_GetError());
	}

	defer ext.SDL_Quit();

	{
		const base_prefix = ext.SDL_GetBasePath() orelse {
			return ext.SDL_LogError(ext.SDL_LOG_CATEGORY_APPLICATION, "%s\n", ext.SDL_GetError());
		};

		defer ext.SDL_free(base_prefix);

		const window_flags = 0;
		const window_pos = ext.SDL_WINDOWPOS_CENTERED;

		const window = ext.SDL_CreateWindow(&app_manifest.title, window_pos, window_pos, app_manifest.width, app_manifest.height, window_flags) orelse {
			return ext.SDL_LogError(ext.SDL_LOG_CATEGORY_APPLICATION, "%s\n", ext.SDL_GetError());
		};

		defer ext.SDL_DestroyWindow(window);

		const renderer_flags = 0;

		const renderer = ext.SDL_CreateRenderer(window, -1, renderer_flags) orelse {
			return ext.SDL_LogError(ext.SDL_LOG_CATEGORY_APPLICATION, "%s\n", ext.SDL_GetError());
		};

		defer ext.SDL_DestroyRenderer(renderer);

		while (true) {
			// TODO: Delta timing.
			var event = @as(ext.SDL_Event, undefined);

			while (ext.SDL_PollEvent(&event) != 0) {
				switch (event.type) {
					ext.SDL_QUIT => return,
					else => {},
				}
			}

			if (ext.SDL_SetRenderDrawColor(renderer, 0, 0, 0, 0) != 0) {
				return ext.SDL_LogError(ext.SDL_LOG_CATEGORY_APPLICATION, "%s\n", ext.SDL_GetError());
			}

			if (ext.SDL_RenderClear(renderer) != 0) {
				return ext.SDL_LogError(ext.SDL_LOG_CATEGORY_APPLICATION, "%s\n", ext.SDL_GetError());
			}

			// TODO: Render here.

			ext.SDL_RenderPresent(renderer);
		}
	}
}
